@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#include "../config.h"
+#include <config.h>
 #include <sys/types.h>
 #include <assert.h>
 #include <ctype.h>
@@ -42,8 +42,9 @@ main(int argc, char **argv)
 	struct _vte_termcap *termcap = NULL;
 	struct _vte_buffer *buffer = NULL;
 	GArray *array;
-	int i, j, l;
-	char c;
+	unsigned int i, j;
+	int l;
+	char b;
 	GValue *value;
 	FILE *infile = NULL;
 	struct _vte_iso2022_state *subst;
@@ -51,7 +52,7 @@ main(int argc, char **argv)
 	GQuark quark;
 	GValueArray *values;
 
-	_vte_debug_parse_string(getenv("VTE_DEBUG_FLAGS"));
+	_vte_debug_init();
 
 	if (argc < 2) {
 		g_print("usage: %s terminal [file]\n", argv[0]);
@@ -83,8 +84,8 @@ main(int argc, char **argv)
 
 	subst = _vte_iso2022_state_new(NULL, NULL, NULL);
 
-	while (fread(&c, 1, 1, infile) == 1) {
-		_vte_buffer_append(buffer, &c, 1);
+	while (fread(&b, 1, 1, infile) == 1) {
+		_vte_buffer_append(buffer, &b, 1);
 	}
 	_vte_iso2022_process(subst, buffer->bytes,
 			_vte_buffer_length(buffer), array);
